/*
 * Copyright (C) 2011 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.

 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Ken VanDine <ken.vandine@canonical.com>
 */

using Dee;
using Gee;
using Config;
using Unity;
using Gwibber;

namespace UnityGwibber {

  /* DBus name for the place. Must match out .place file */
  const string BUS_NAME = "com.canonical.Unity.Lens.Gwibber";

  /**
   * The Daemon class implements all of the logic for the place.
   *
   */
  public class Daemon : GLib.Object
  {
    private Unity.Lens lens;
    private Unity.Scope scope;
    private Gwibber.Streams streams_service;
    private Gwibber.Service service;
    private Gwibber.Accounts accounts;
    private Gwibber.Utils utils;
    private Dee.Model? _model;
    private Dee.Model? _streams_model;
    private Dee.Filter _sort_filter;
    /* Keep track of the previous search, so we can determine when to
     * filter down the result set instead of rebuilding it */
    private unowned Dee.ModelIter _stream_iter_first = null;
    private unowned Dee.ModelIter _stream_iter_last = null;


    construct
    {
      streams_service = new Gwibber.Streams();
      service = new Gwibber.Service();
      utils = new Gwibber.Utils();
      accounts = new Gwibber.Accounts();

      scope = new Unity.Scope ("/com/canonical/unity/scope/gwibber");
      scope.search_in_global = true;

      lens = new Unity.Lens("/com/canonical/unity/lens/gwibber", "gwibber");
      lens.search_in_global = true;
      lens.search_hint = _("Enter name or content you would like to search for");
      lens.visible = true;
      populate_categories ();
      populate_filters();
      lens.add_local_scope (scope);

      /* Listen for filter changes */
      scope.notify["active"].connect(() =>
      {
        if (scope.active)
        {
          scope.queue_search_changed (SearchType.DEFAULT);
        }
      });

      scope.search_changed.connect ((lens_search, search_type, cancellable) =>
      {
        if (search_type == SearchType.DEFAULT)
          update_scope_search.begin (lens_search, cancellable);
        else
          update_global_search.begin (lens_search, cancellable);
      });

      lens.notify["active"].connect ((obj, pspec) => {
        if (lens.active && scope.active)
        {
          if (_stream_iter_first != _model.get_first_iter () || _stream_iter_last != _model.get_last_iter ())
          {
            if (scope.active)
            {
              scope.queue_search_changed (SearchType.DEFAULT);
            }
          }
        }
      });

      scope.filters_changed.connect (() =>
      {
        scope.queue_search_changed (SearchType.DEFAULT);
      });

      try
      {
        lens.export ();
      } catch (GLib.IOError e)
      {
        warning ("failed to export lens: %s", e.message);
      }

      _streams_model = streams_service.stream_model;
      Intl.setlocale(LocaleCategory.COLLATE, "C");
      _sort_filter = Dee.Filter.new_collator_desc (StreamModelColumn.TIMESTAMP);
      _model = new Dee.FilterModel (_streams_model, _sort_filter);
    }

    private void populate_filters ()
    {
      var filters = new GLib.List<Unity.Filter> ();

      /* Stream filter */
      {
        var filter = new CheckOptionFilter ("stream", _("Stream"));

        filter.add_option ("messages", _("Messages"));
        filter.add_option ("replies", _("Replies"));
        filter.add_option ("images", _("Images"));
        filter.add_option ("videos", _("Videos"));
        filter.add_option ("links", _("Links"));
        filter.add_option ("private", _("Private"));
        filter.add_option ("public", _("Public"));

        filters.append (filter);
      }

      /* Account filter */
      {
        var filter = create_account_filter ();
        filters.append (filter);
      }

      lens.filters = filters;

      accounts.created.connect((source) => {
        /* FIXME: we need a way to add an option 
        var filter = scope.get_filter ("account_id") as CheckOptionFilter;
        filter.add_option (source.id, source.service + "/" + source.username);
        */
      });

      accounts.deleted.connect(() => {
        /* FIXME: we need a way to remove an option or remove and re-add the filter 
        var filter = scope.get_filter ("account_id") as CheckOptionFilter;
        lens.filters.remove (filter);
        filter = null;
        filter = create_account_filter ();
        lens.filters.append (filter);
        */
      });
    }

    private Unity.CheckOptionFilter create_account_filter ()
    {
      GLib.Icon icon;
      GLib.File icon_file;
      var filter = new CheckOptionFilter ("account_id", _("Account"));
      foreach (var _acct in accounts.list ())
      {
        /* create a service icon, which isn't used yet as of libunity 4.24
           when it is used, we can drop _acct.service from the option name 
           and rely on the service icon and username to disambiguate accounts */
        icon_file = GLib.File.new_for_path (GLib.Path.build_filename (Config.PKGDATADIR + "/plugins/" + _acct.service + "/ui/icons/16x16/" + _acct.service + ".png"));
        icon = new GLib.FileIcon (icon_file);

        if (_acct.name != null)
          filter.add_option (_acct.id, _acct.name, icon);
        else
          filter.add_option (_acct.id, _acct.service + "/" + _acct.username, icon);
      }
      return filter;
    }


    private void populate_categories ()
    {
      var categories = new GLib.List<Unity.Category> ();
      Icon icon;

      icon = new ThemedIcon ("tray-message");
      var cat = new Unity.Category (_("Messages"), icon, Unity.CategoryRenderer.HORIZONTAL_TILE);
      categories.append (cat);

      icon = new ThemedIcon ("mail_reply");
      cat =  new Unity.Category (_("Replies"), icon, Unity.CategoryRenderer.HORIZONTAL_TILE);
      categories.append (cat);

      icon = new ThemedIcon ("gnome-mime-image");
      cat =  new Unity.Category (_("Images"), icon, Unity.CategoryRenderer.HORIZONTAL_TILE);
      categories.append (cat);

      icon = new ThemedIcon ("video");
      cat =  new Unity.Category (_("Videos"), icon, Unity.CategoryRenderer.HORIZONTAL_TILE);
      categories.append (cat);

      icon = new ThemedIcon ("web-browser");
      cat =  new Unity.Category (_("Links"), icon, Unity.CategoryRenderer.HORIZONTAL_TILE);
      categories.append (cat);

      icon = new ThemedIcon ("mail-send-receive");
      cat =  new Unity.Category (_("Private"), icon, Unity.CategoryRenderer.HORIZONTAL_TILE);
      categories.append (cat);

      icon = new ThemedIcon ("evolution");
      cat =  new Unity.Category (_("Public"), icon, Unity.CategoryRenderer.HORIZONTAL_TILE);
      categories.append (cat);

      lens.categories = categories;
    }


    private bool search_is_invalid (LensSearch? search)
    {
      /* This boolean expression is unfolded as we seem to get
       * some null dereference if we join them in a big || expression */
      if (search == null)
        return true;
      else if (search.search_string == null)
        return true;

      return search.search_string.strip() == "";
    }

    private string prepare_search_string (LensSearch? search)
    {
      var s = search.search_string;

      if (s.has_suffix (" "))
        s = s.strip ();

      return s;
    }


    private async void update_global_search (LensSearch search, Cancellable cancellable)
    {
      var results_model = scope.global_results_model;

      if (search_is_invalid (search))
        {
          return;
        }

      var search_string = prepare_search_string (search);

      update_results_model (results_model, search_string, null);

      search.finished ();
    }

    private async void update_scope_search  (LensSearch search, Cancellable cancellable)
    {
      var results_model = search.results_model;

      var search_string = prepare_search_string (search);

      update_results_model (results_model, search_string, null);

      search.finished ();
    }


    /* Generic method to update a results model. We do it like this to minimize
     * code dup between updating the global- and the entry results model */
    private void update_results_model (Dee.Model results_model,
                                       string? search, Categories? category)
    {
      unowned Dee.ModelIter iter, end;
      string search_string = null;
      Categories group = Categories.MESSAGES;

      var stream_ids = new Gee.ArrayList<string> ();
      var filter = scope.get_filter("stream") as CheckOptionFilter;
      foreach (Unity.FilterOption option in filter.options)
      {
        if (option.active)
        {
          stream_ids.add (option.id);
        }
      }

      var account_ids = new Gee.ArrayList<string> ();
      filter = scope.get_filter("account_id") as CheckOptionFilter;
      foreach (Unity.FilterOption option in filter.options)
      {
        if (option.active)
        {
          account_ids.add (option.id);
        }
      }

      results_model.clear ();

      iter = _model.get_first_iter ();
      end = _model.get_last_iter ();

      _stream_iter_first = _model.get_first_iter ();
      _stream_iter_last = end;
      
      if (search != null)
      { 
          search_string = search.down();
      }

      while (iter != end)
      {
        bool stream_match = true;
        bool account_match = true;
        if (stream_ids != null)
        {
          if (stream_ids.size > 0)
          {
            if (!(_model.get_string(iter, StreamModelColumn.STREAM) in stream_ids))
            {
              stream_match = false;
            }
          }
        }
        if (account_ids.size > 0)
        {
          string[] _accounts_array = (string[])_model.get_value (iter, StreamModelColumn.ACCOUNTS);
          foreach (var a in _accounts_array)
          {
            string _account = a.split(":")[0];
            if (!(_account in account_ids))
              account_match = false;
          }
        }
        if (!(account_match && stream_match))
        {
          iter = _model.next (iter);
          continue;
        }

        if (search_string == null || search_string in _model.get_string(iter, StreamModelColumn.SENDER).down() || search_string in _model.get_string(iter, StreamModelColumn.MESSAGE).down())
        {
          if (_model.get_string(iter, StreamModelColumn.STREAM) == "messages")
            group = Categories.MESSAGES;
          else if (_model.get_string(iter, StreamModelColumn.STREAM) == "replies")
            group = Categories.REPLIES;
          else if (_model.get_string(iter, StreamModelColumn.STREAM) == "images")
            group = Categories.IMAGES;
          else if (_model.get_string(iter, StreamModelColumn.STREAM) == "videos")
            group = Categories.VIDEOS;
          else if (_model.get_string(iter, StreamModelColumn.STREAM) == "links")
            group = Categories.LINKS;
          else if (_model.get_string(iter, StreamModelColumn.STREAM) == "private")
            group = Categories.PRIVATE;
          else if (_model.get_string(iter, StreamModelColumn.STREAM) == "public")
            group = Categories.PUBLIC;
          
          string _icon_uri = _model.get_string(iter, StreamModelColumn.ICON_URI);
          var _avatar_cache_image = utils.avatar_path(_icon_uri);
          if (_avatar_cache_image == null)
          {
            try
            {
              _avatar_cache_image = service.avatar_path (_icon_uri);
            } catch (GLib.Error e)
            {
            }
            if (_avatar_cache_image == null)
              _avatar_cache_image = _icon_uri;
          }

          results_model.append(_model.get_string(iter, StreamModelColumn.URL), _avatar_cache_image, group, "text/html", _model.get_string(iter, StreamModelColumn.SENDER), _model.get_string(iter, StreamModelColumn.MESSAGE));
        }
        if (iter == end)
          iter = null;
        else
          iter = _model.next (iter);
      }
     
      //debug ("Results has %u rows", results_model.get_n_rows());
    }

  } /* End Daemon class */
} /* end Gwibber namespace */
